==> nono-0.4.4.tar.gz: nono-0.4.4/debugger/debugger_private.h <==
//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// デバッガ (内部用)
//

#pragma once

#include "branchhistory.h"
#include "bus.h"
#include "debugger_memory.h"
#include "disasm.h"
#include "memdump.h"
#include "monitor.h"
#include "textscreen.h"
#include "thread.h"
#include <array>
#include <condition_variable>
#include <mutex>

using ConstStringPair = std::pair<const std::string, const std::string>;
using HelpMessages = std::vector<ConstStringPair>;

class HostCOMDevice;
class VectorTable;

class Debugger;

// デバッガの CPU 依存部分
class DebuggerMD
{
 public:
	enum class Arch {
		M680x0,
		M88xx0,
	};

 public:
	DebuggerMD(Debugger *parent_, Arch arch_) {
		parent = parent_;
		arch = arch_;
	}
	virtual ~DebuggerMD();

	// 特権モードなら true を返す
	virtual bool IsSuper() const = 0;
	// MMU が有効なら true を返す
	virtual bool MMUEnabled() const = 0;

	// アドレス変換をする。
	// MMUEnabled() が true であることは呼び出し側が確認すること。
	// lookup が true ならテーブルサーチまで行う。
	// 変換できなければ (uint64)-1 を返す。
	virtual uint64 TranslateAddr(saddr_t laddr, bool lookup) const = 0;

	// 現在の命令先頭アドレスを返す
	virtual uint32 GetPC() const = 0;

	// ステップアウトを設定する
	virtual void SetStepOut() = 0;
	// ステップアウトなら true を返す
	virtual bool IsStepOut() const = 0;

	// この命令がステップイン出来るなら true を返す (cmd_n 用)
	virtual bool IsOpStepIn(DebuggerMemoryStream& mem) = 0;

	// name で示されるのレジスタの値を返す。
	// メモリダンプで指定するケースを想定しているので、明らかにアドレスでは
	// ないレジスタは含まなくてよい (m680x0 の SR など)。ただし Dn は含めて
	// よい。また m680x0 の SRP/CRP は下位32ビットでよい。
	// name が正しくない場合は (uint64)-1 を返す。
	virtual uint64 GetRegAddr(const char *name) const = 0;

	// 現在のレジスタセットを内部にバックアップする
	virtual void BackupRegs() = 0;

	// レジスタ表示系のコマンドを実行。処理すれば true を返す。
	virtual bool ShowRegister(FILE *, const std::vector<std::string>& args) = 0;
	// レジスタ表示系コマンドのヘルプ一覧を取得
	virtual const HelpMessages& GetHelpListReg() const = 0;
	// レジスタ表示系コマンドのヘルプメッセージを取得
	virtual const HelpMessages& GetHelpReg() const = 0;

	// ブランチ履歴・例外履歴(オブジェクト)を取得
	virtual BranchHistory& GetBrHist() const = 0;
	virtual BranchHistory& GetExHist() const = 0;

	// 直近の命令でアクセスした laddr が一致したら true を返す
	// 1 命令で複数にアクセスする場合を考慮して、md 側で判断する
	virtual bool CheckLEA(uint32 laddr) = 0;

	// mem.laddr の位置の命令を一つ逆アセンブルする。
	virtual std::string Disassemble(DebuggerMemoryStream& mem) = 0;

	// 機種
	Arch arch {};

	// 最短の1命令のバイト数
	uint inst_bytes {};
	// 命令が固定長ならそのバイト数、可変長なら 0
	uint inst_bytes_fixed {};

 protected:
	Debugger *parent {};	// 親クラス
};

class Debugger : public ThreadDevice
{
	using inherited = ThreadDevice;
	friend class DebuggerMD_m680x0;
	friend class DebuggerMD_m88xx0;

 private:
	// リクエストフラグ
	static const uint32 REQUEST_EXIT	= 0x0001;	// 終了
	static const uint32 REQUEST_RXCHAR	= 0x0002;	// 1文字受信
	static const uint32 REQUEST_ACCEPT	= 0x0004;	// 着信通知
	static const uint32 REQUEST_PROMPT	= 0x0008;	// プロンプト要求

	// コマンド
	enum CmdAct {
		Stay,		// デバッガプロンプトに留まる
		Leave,		// デバッガプロンプトを抜けて VM を実行する
		Quit,		// デバッガとの接続を終了する
	};
	struct cmddef_t {
		const char *name;
		CmdAct (Debugger::*func)();
	};

	// ブレークポイント種別
	enum BreakpointType {
		Unused = 0,
		Address,		// このアドレスに来る直前
		Memory,			// このアドレスをアクセスした直後
		Exception,		// この例外を検出した直後
		Instruction,	// この命令を実行する直前
	};

	// ブレークポイント
	struct breakpoint_t {
		BreakpointType type {};	// 種別
		// 種別ごとのパラメータ
		union {
			uint32 addr {};		// ターゲットアドレス (Address/Memory)
			struct {
				int vec1;		// ベクタ番号 (開始)
				int vec2;		// ベクタ番号 (終了)
			};
			struct {
				uint32 inst;	// 命令ワード
				uint32 mask;	// マスク
			};
		};

		uint32 matched {};		// 成立回数 (積算)

		// skip はユーザ指定値。ただし
		// -1 なら常にスキップ、つまりブレークせずカウントのみ行う、
		// 0 ならスキップなし、つまり成立ごとにブレーク(これがデフォルト)、
		// n(>0) なら n 回の成立をスキップし、(n+1) 回目でブレークの意。
		int32 skip {};			// スキップ回数 (ユーザ指定値)
		int32 skipremain {};	// 残りスキップ回数 (0 で成立)
	};
	static const int MAX_BREAKPOINTS = 8;

 public:
	Debugger();
	~Debugger() override;

	void SetLogLevel(int) override;
	bool Create() override;
	bool Init() override;
	void ThreadRun() override;
	void Terminate() override;

	void Exec();
	void NotifyException(int vector);

	// MPU トレースが必要なら true を返す
	bool IsTrace() const;

	// FILE コールバック
	int ReadFunc(char *, int);
	int WriteFunc(const char *, int);

	DebuggerMD *GetMD() const { return md.get(); }

 private:
	void Close();
	void RxCallback();
	void AcceptCallback();
	void Input(int);
	void EnterPrompt();
	void LeavePrompt(bool trace);
	void PrintPrompt();

	CmdAct Command();
	bool CheckAllBreakpoints();
	bool CheckBreakpointInst(breakpoint_t&);
	void ParseCmdbuf();
	// コマンド名は大文字小文字を区別する関係でスネークスタイル。
	CmdAct cmd_b();
	CmdAct cmd_bi();
	CmdAct cmd_bm();
	CmdAct cmd_bv();
	CmdAct cmd_b_list();
	CmdAct cmd_b_set(BreakpointType);
	CmdAct cmd_b_delete();
	CmdAct cmd_brhist();
	CmdAct cmd_bx();
	CmdAct cmd_c();
	CmdAct cmd_ct();
	CmdAct cmd_d();
	CmdAct cmd_dt();
	CmdAct cmd_D();
	CmdAct cmd_disp();
	CmdAct cmd_exhist();
	CmdAct cmd_h();
	CmdAct cmd_hb();
	CmdAct cmd_hr();
	CmdAct cmd_L();
	CmdAct cmd_m();
	CmdAct cmd_mt();
	CmdAct cmd_M();
	CmdAct cmd_minus();
	CmdAct cmd_n();
	CmdAct cmd_nt();
	CmdAct cmd_q();
	CmdAct cmd_reset();
	CmdAct cmd_s();
	CmdAct cmd_st();
	CmdAct cmd_so();
	CmdAct cmd_sot();
	CmdAct cmd_show();
	CmdAct cmd_t();

	bool Check();
	int AddBreakpoint(const breakpoint_t&);
	void RecalcInstMask();
	bool GetAddr(saddr_t *addr, bool);
	CmdAct cmd_d_common(MemoryMode access_mode, MMULookupMode lookup_mode);
	CmdAct cmd_hist_common(BranchHistory& hist);
	CmdAct cmd_m_common(MemoryMode access_mode, MMULookupMode lookup_mode);
	CmdAct cmd_n_common(bool);
	CmdAct cmd_s_common(bool);
	CmdAct cmd_so_common(bool);
	bool ParseVerbHex(const char *arg, uint32 *valp);
	bool ParseAddr(const char *arg, uint32_t *addrp);
	bool ParseTime(const char *arg, uint64 *timep);
	void SetNBreakpoint();

	static const HelpMessages HelpListMain;
	static const HelpMessages HelpListBreakpoints;
	static const HelpMessages HelpDetails;
	CmdAct ShowHelpList(const HelpMessages& msgs);

	// 個別ヘルプメッセージを出力用に置換
	std::string HelpConvert(const std::string& msg);

	// モニタを更新して表示
	void ShowMonitor(Monitor& monitor);
	// テキストスクリーンを表示
	void ShowTextScreen(TextScreen& screen);

	std::unique_ptr<DebuggerMD> md {};	// 機種依存部分

	std::mutex mtx {};
	// このスレッドへのリクエスト
	std::condition_variable cv_request {};
	uint32 request {};
	// コマンドモード(プロンプト)なら true
	std::condition_variable cv_prompt {};
	bool prompt_released {};

	bool is_prompt {};

	// コンソール
	std::unique_ptr<HostCOMDevice> hostcom /*{}*/;
	FILE *cons {};

	const char *prompt {};		// プロンプト文字列
	std::string cmdbuf {};		// 現在のコマンドライン
	std::string last_cmdbuf {};	// 直前のコマンドライン
	std::vector<std::string> args {};

	uint32 pc {};
	std::array<breakpoint_t, MAX_BREAKPOINTS> bpoint {};
	bool bc_enable {};
	uint32 bc_addr {};
	uint32 bi_inst {};
	int bi_inst_bytes {};
	int bi_need_bytes {};
	int bv_vector = -1;
	bool ct_enable {};
	uint64 ct_end_time {};
	uint64 ct_timespan {};
	saddr_t d_last_addr {};
	saddr_t m_last_addr {};
	bool n_enable {};
	uint32 n_breakaddr {};
	uint32 n_count {};
	bool s_enable {};
	uint32 s_count {};
	bool so_enable {};
	bool t_enable {};
	uint32 t_count {};

	// 一度VMを実行して再びプロンプトに来たら true
	bool is_continued {};

	// MPU を一時停止するとき true (ワンショット)
	bool is_pause {};

	// プロンプトで表示するレジスタ群
	std::vector<std::string> disp_regs {};

	// ブレークポイント到達メッセージ
	// (コンソールが獲得できるまで保留しておくため)
	std::string bpointmsg {};

	// ブレークポイントモニタ
	DECLARE_MONITOR_CALLBACK(MonitorUpdateBpoint);
	Monitor bpoint_monitor { this };

	// 逆アセンブルモニタ
	std::array<std::unique_ptr<DisasmMonitor>, MAX_DISASM_MONITOR>
		disasm_monitor {};

	// メモリダンプモニタ
	std::array<std::unique_ptr<MemdumpMonitor>, MAX_MEMDUMP_MONITOR>
		memdump_monitor {};

	// ベクタテーブル
	std::unique_ptr<VectorTable> pVectorTable /*{}*/;

	static std::vector<cmddef_t> cmdtable;
};

extern Debugger *gDebugger;

#define NORM		"\x1b[0m"
#define BOLD		"\x1b[1m"
#define BOLDIF(n)	((n) ? BOLD : "")

==> nono-0.6.3.tar.gz: nono-0.6.3/debugger/debugger_private.h <==
//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// デバッガ (内部用)
//

#pragma once

#include "debugger_defs.h"
#include "fcaddr.h"

class IODevice;

enum class CPUState
{
	Normal = 0,
	Stop,		// STOP 命令など割り込み待ち状態
	Halt,		// ホールト、リセットなど非動作状態
};

// デバッガの CPU 依存部分
class DebuggerMD
{
 public:
	DebuggerMD(Debugger *parent_, CPUArch arch_);
	virtual ~DebuggerMD();

	// CPU の実行状態を返す。
	virtual CPUState GetCPUState() const = 0;
	// 特権モードなら true を返す
	virtual bool IsSuper() const = 0;
	// MMU が有効なら true を返す
	virtual bool MMUEnabled() const = 0;

	// アドレス変換をする。
	// MMUEnabled() が true であることは呼び出し側が確認すること。
	// lookup が true ならテーブルサーチまで行う。
	// 変換できなければ (uint64)-1 を返す。
	virtual uint64 TranslateAddr(fcaddr laddr, bool lookup) const = 0;

	// 現在の命令先頭アドレスを返す
	virtual uint32 GetPC() const = 0;

	// ステップアウトを設定する
	virtual void SetStepOut() = 0;
	// ステップアウトなら true を返す
	virtual bool IsStepOut() const = 0;

	// この命令がステップイン出来るなら true を返す (cmd_n 用)
	virtual bool IsOpStepIn(DebuggerMemoryStream& mem) = 0;

	// name で示されるのレジスタの値を返す。
	// メモリダンプで指定するケースを想定しているので、明らかにアドレスでは
	// ないレジスタは含まなくてよい (m680x0 の SR など)。ただし Dn は含めて
	// よい。また m680x0 の SRP/CRP は下位32ビットでよい。
	// name が正しくない場合は (uint64)-1 を返す。
	virtual uint64 GetRegAddr(const char *name) const = 0;

	// 現在のレジスタセットを内部にバックアップする
	virtual void BackupRegs() = 0;

	// レジスタ表示系のコマンドを実行。処理すれば true を返す。
	virtual bool ShowRegister(FILE *, const std::vector<std::string>& args) = 0;
	// レジスタ表示系コマンドのヘルプ一覧を取得
	virtual const HelpMessages& GetHelpListReg() const = 0;
	// レジスタ表示系コマンドのヘルプメッセージを取得
	virtual const HelpMessages& GetHelpReg() const = 0;

	// 直近の命令でアクセスした laddr が一致したら true を返す
	// 1 命令で複数にアクセスする場合を考慮して、md 側で判断する
	virtual bool CheckLEA(uint32 laddr) = 0;

	// mem.laddr の位置の命令を一つ逆アセンブルする。
	virtual std::string Disassemble(DebuggerMemoryStream& mem) = 0;

	// 機種名を取得する
	const std::string& GetName() const { return name; }

	// この CPU のバスを返す
	IODevice *GetBus() const { return bus; }
	// このバスの論理アドレス幅 (ビット数) を返す
	int GetLASize() const { return lasize; }
	// このバスの物理アドレス幅 (ビット数) を返す
	int GetPASize() const { return pasize; }

	// 機種
	CPUArch arch {};

	// 最短の1命令のバイト数
	uint inst_bytes {};
	// 命令が固定長ならそのバイト数、可変長なら 0
	uint inst_bytes_fixed {};

	// 例外ブレーク。
	// 機種固有ではないが main と xp でそれぞれに必要なので、ここに持っておく。
	int bv_vector {};

 protected:
	Debugger *parent {};	// 親クラス

	std::string name {};	// 機種名

	IODevice *bus {};		// 対応する外部バス
	int lasize {};			// バスの論理アドレス幅
	int pasize {};			// バスの物理アドレス幅
};

#define NORM		"\x1b[0m"
#define BOLD		"\x1b[1m"
#define BOLDIF(n)	((n) ? BOLD : "")
